from lovely.pyrest.service import Service

svc = Service(name='svc', path='/svc')

schema = {
    "query": {
        "type": "object",
        "properties": {
            "id": {
                "type": "string",
                "required": False
            }
        }
    }
}

@svc.get(schema=schema)
def get(request):
    return {'title': 'Hello World!'}



from lovely.pyrest.service import Service

articles = Service(name='article', path='/article/{id}')

ARTICLES = {}


def article_exists(request):
    """ A Validator the verify that the article exists """
    _id = request.matchdict['id']
    if _id not in ARTICLES:
        desc = "Article with id %s does not exist" % _id
        request.errors.add('uri', desc)  # Add an error if validation fails
        request.errors.status = 404  # Set the error status code


@articles.get(validators=[article_exists])
def get(request):
    """ Returns an article """
    return ARTICLES.get(request.matchdict['id'])


# Create a schema to validate the body
schema = {
    "body": {
        "type": "object",
        "properties": {
            "title": {
                "type": "string"
            },
            "content": {
                "type": "string",
                "required": False
            },
        }
    }
}


@articles.post(schema=schema)
def post(request):
    """ Save an article """
    _id = request.matchdict['id']
    return {'stored': _id}



from lovely.pyrest.service import Service

article = Service(name='article', path='/article')

schema = {
    "body": {
        "type": "object",
        "properties": {
            "title": {
                "type": "string"
            }
        }
    },
    "query": {
        "type": "object",
        "properties": {
            "id": {
                "type": "string"
            },
            "categories": {
                "type": "array",
                "required": False,
                "items": {
                    "enum": ["news", "sports", "politics"]
                }
            }
        }
    }
}


@article.post(schema=schema)
def post(request):
    """ Save an article """
    return {'stored': 'ok'}


@article.get(schema=schema)
def get(request):
    """ Save an article """
    return {'stored': 'ok'}



from lovely.pyrest.predicates import ContentTypePredicate
from lovely.pyrest.views import get_fallback_view, decorate_view
from pyramid.events import NewRequest
from errors import Errors
from pyramid.renderers import JSONP
import copy

__version__ = "0.1.2"


def add_service(config, service):
    """ registers the views for a service """
    # register the route
    config.add_route(service.name, service.path)
    # create the fallback view
    config.add_view(view=get_fallback_view(service),
                    route_name=service.name)
    for method, view, args in service.definitions:
        args = copy.deepcopy(args)  # make a copy of the dict to not modify it
        args['request_method'] = method
        decorated_view = decorate_view(view, dict(args))
        # remove args which are unknown by pyramid
        for item in ('validators', 'schema', 'jsonp', 'help'):
            if item in args:
                del args[item]
        config.add_view(view=decorated_view, route_name=service.name, **args)
    config.commit()


def wrap_request(event):
    """ Creates an empty errors list and adds it to the request """
    request = event.request
    if not hasattr(request, 'errors'):
        request.errors = Errors()


# The includeme function will be called after including
# this module:
# config.include('lovely.pyrest')
def includeme(config):
    config.add_directive('add_service', add_service)
    config.add_subscriber(wrap_request, NewRequest)
    config.add_view_predicate('content_type', ContentTypePredicate)
    add_custom_config(config)


def add_custom_config(config):
    """ Custom configuration parameters in your .ini file
    will be handled here.
    """
    settings = config.get_settings()

    # Add JSONP support
    # If param_name is set as qeury parameter in a request it will trigger
    # the JSONP transformation.
    # 'callback' will be set as default if `param_name` is not specified.
    param_name = settings.get('lovely.pyrest.jsonp.param_name', 'callback')
    config.add_renderer('jsonp', JSONP(param_name=param_name))



class ContentTypePredicate(object):
    """
    Predicate for matching the `Content-type` request header.
    See: http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/hooks.html#view-and-route-predicates
    """
    def __init__(self, val, config):
        self.val = val

    # The result of phash is not seen in output anywhere, it just informs the
    # uniqueness constraints for view configuration.
    def text(self):
        return 'content_type = %s' % (self.val,)

    phash = text

    def __call__(self, context, request):
        return request.content_type == self.val



from lovely.pyrest.service import Service

article = Service(name='article', path='/article')


def validate_query(request):
    """ Check if the id field is set """
    if not 'id' in request.GET:
        request.errors.add('query',
                           'Parameter `id` is missing')
        request.errors.status = 400


@article.get(validators=[validate_query])
def get(request):
    return {'status': 'success'}



from docutils.parsers.rst import Directive
from docutils import nodes
from importlib import import_module
from lovely.pyrest.service import get_service
from lovely.pyrest.sphinx.helpers import create_node, trim, empty_node
import json


class SchemaDirective(Directive):
    """ The Schema directive renders the information about GET-Parameter and
        JSON-Body of a specific service. """

    has_content = True
    required_arguments = 3

    def run(self):
        module, svc, method = self.arguments
        import_module(module)
        service = get_service(svc)
        env = self.state.document.settings.env
        service_id = "service_%d" % env.new_serialno('service')
        return [SchemaDirective.render_schema(service,
                                              method,
                                              service_id)]

    @staticmethod
    def render_schema(service, method, service_id):
        """ Renders schema information of a service """
        # check if service is not null
        if not service:
            return empty_node()
        schema = service.get_argument('schema', method)
        # check that the schema exists
        if not schema:
            return empty_node()
        schema_id = "%s_%s_%s" % (service_id,
                                  method,
                                  'schema')
        node = nodes.section(ids=[schema_id])
        # if the schema contains a `query` schema render GET-Parameters
        # documentation:
        # <name>: <required>, <type>, <description>, <other arguments>
        if 'query' in schema and 'properties' in schema['query']:
            title = nodes.title(text='GET-Parameters:')
            node +=title
            properties = schema['query']['properties']
            for param, spec in properties.iteritems():
                prop_node = nodes.list_item()
                desc = "%s: " % param
                extras = []
                # Check if required is true
                # and add (required | optional)
                if spec.get('required', True):
                    extras.append('required')
                else:
                    extras.append('optional')
                # Check if the type is set and add it as second
                # parameter
                if 'type' in spec:
                    extras.append('' + spec['type'])
                # Check if the description is set and add it as
                # third parameter
                if 'description' in spec:
                    extras.append('' + spec['description'])

                # add the other fields
                for k, v in spec.iteritems():
                    if k in ['type', 'description', 'required']:
                        continue
                    extras.append(' %s: %s' % (k, v))
                desc += ', '.join(extras)
                prop_node += nodes.inline(text=desc)

                node += prop_node
        # if the schema has a `body` schema convert it to string and
        # insert it into the documentation
        if 'body' in schema:
            body_schema = schema['body']
            title = nodes.title(text='JSON Body:')
            text = json.dumps(body_schema, indent=4)
            # prefix every line with a pipe, so the rst conversation returns a
            # line_block where the spaces are preserved
            text = '\n'.join(['| ' + l for l in text.splitlines()])
            node += title
            node += create_node(trim(text))
        return node



from lovely.pyrest.service import Service

""" This module contains several service definitions with
    predicates and validators for integration testing """


# Define a simple service
service = Service('myService', '/service')


@service.get()
def get(request):
    return {'test': 'succeed'}


# Define a more complex service with validators
# and predicates
user = Service('user', '/user')


@user.get(accept='application/json',
          content_type='application/json')
def get_user(request):
    return {'test': 'succeed'}


schema = {
    "query": {
        "type": "object",
        "properties": {
            "u": {
                "type": "string",
                "format": "url"
            }
        }
    },
    "body": {
        "type": "object",
        "properties": {
            "name": {
                "type": "string"
            }
        }
    }
}


@user.post(schema=schema)
def post(request):
    return {'post_test': 'succeeded'}



import validictory


def number(s):
    try:
        return int(s)
    except ValueError:
        return float(s)

CONVERSION = {
    "integer": lambda s: int(s),
    "number": number,
    "boolean": lambda s: True if s.lower() == "true" else False,
    "array": lambda s: s.split(","),
}


def validate_schema(request, schema):
    # validate schema
    query_schema = schema.get("query")
    if query_schema is not None:
        try:
            # Because all GET-Parameters are strings try to convert them
            # into the format specified in the schema
            params = {}
            for k in request.GET:
                params[k] = request.GET[k]
            for prop, o in query_schema.get('properties').iteritems():
                stype = o.get('type')
                if stype in CONVERSION:
                    converter = CONVERSION[stype]
                    v = params.get(prop)
                    if v is not None:
                        params[prop] = converter(v)
            # for convenience remember the params_dict on the request
            request.params_dict = params
            validictory.validate(params, query_schema)
        except ValueError, error:
            request.errors.add('query',
                               error.message)
    body_schema = schema.get("body")
    if body_schema is not None:
        try:
            json_body = request.json_body
            validictory.validate(json_body, body_schema)
        except ValueError, error:
            request.errors.add('body',
                               error.message)
